/**
 * Copyright (c) 2021 Dominic Msters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "shaderprogram.h"

void shaderInit(shaderprogram_t *shader,
  char *vertexShaderSource, char* fragmentShaderSource
) {
  int isSuccess, maxLength;
  char *error;
  GLuint shaderVertex, shaderFragment, shaderProgram;

  ASSERT_NOT_NULL(shader);
  ASSERT_NOT_NULL(vertexShaderSource);
  ASSERT_NOT_NULL(fragmentShaderSource);

  // Load the vertex shader first
  shaderVertex = glCreateShader(GL_VERTEX_SHADER);
  glShaderSource(shaderVertex, 1, &vertexShaderSource, 0);
  glCompileShader(shaderVertex);

  // Validate
  glGetShaderiv(shaderVertex, GL_COMPILE_STATUS, &isSuccess);
  if(!isSuccess) {
    glGetShaderiv(shaderVertex, GL_INFO_LOG_LENGTH, &maxLength);
    error = malloc(maxLength);
    glGetShaderInfoLog(shaderVertex, maxLength, &maxLength, error);
    printf("Failed to compile vertex shader %s\n", error);
    free(error);
    return;
  }

  // Now load the Frag shader
  shaderFragment = glCreateShader(GL_FRAGMENT_SHADER);
  glShaderSource(shaderFragment, 1, &fragmentShaderSource, 0);
  glCompileShader(shaderFragment);
  glGetShaderiv(shaderFragment, GL_COMPILE_STATUS, &isSuccess);
  if(!isSuccess) {
    glGetShaderiv(shaderFragment, GL_INFO_LOG_LENGTH, &maxLength);
    error = malloc(maxLength);
    glGetShaderInfoLog(shaderFragment, maxLength, &maxLength, error);
    printf("Failed to compile fragment shader %s\n", error);
    free(error);
    glDeleteShader(shaderVertex);
    return;
  }

  // Now create the shader program.
  shaderProgram = glCreateProgram();
  glAttachShader(shaderProgram, shaderVertex);
  glAttachShader(shaderProgram, shaderFragment);

  //Bind, Verify & Use the shader program
  glLinkProgram(shaderProgram);
  glGetProgramiv(shaderProgram, GL_LINK_STATUS, &isSuccess);
  if(!isSuccess) {
    glGetProgramiv(shaderProgram, GL_INFO_LOG_LENGTH, &maxLength);
    error = malloc(maxLength);
    glGetProgramInfoLog(shaderProgram, maxLength, &maxLength, error);
    printf("Failed to load shader program %s\n", error);
    free(error);
    glDeleteShader(shaderVertex);
    glDeleteShader(shaderFragment);
    return;
  }

  // Everything is okay, let's create the encapsulated shader.
  shader->shaderVertex = shaderVertex;
  shader->shaderFrag = shaderFragment;
  shader->shaderProgram = shaderProgram;
  
  // Bind the shader
  shaderUse(shader);
}

void shaderDispose(shaderprogram_t *shader) {
  ASSERT_NOT_NULL(shader);

  glDeleteProgram(shader->shaderProgram);
  glDeleteShader(shader->shaderVertex);
  glDeleteShader(shader->shaderFrag);
}

void shaderUse(shaderprogram_t *shader) {
  ASSERT_NOT_NULL(shader);
  glUseProgram(shader->shaderProgram);
}

void shaderUseCamera(
  shaderprogram_t *shader, shaderuniform_t uniformView,
  shaderuniform_t uniformProjection, camera_t *camera
) {
  ASSERT_NOT_NULL(shader);
  ASSERT_NOT_NULL(camera);
  ASSERT_GREATER_THAN_EQUAL_TO(uniformView, 0);
  ASSERT_GREATER_THAN_EQUAL_TO(uniformProjection, 0);

  shaderUseMatrix(shader, uniformView, &camera->view);
  shaderUseMatrix(shader, uniformProjection, &camera->projection);
}

void shaderUseTexture(
  shaderprogram_t *shader, shaderuniform_t uniform, textureslot_t slot
) {
  ASSERT_NOT_NULL(shader);
  ASSERT_GREATER_THAN_EQUAL_TO(uniform, 0);
  ASSERT_GREATER_THAN_EQUAL_TO(slot, 0);
  ASSERT_LESS_THAN(slot, TEXTURE_SLOTS_MAX);
  glUniform1i(uniform, slot);
}

void shaderUseMatrix(
  shaderprogram_t *shader, shaderuniform_t uniform, matrix_t *matrix
) {
  ASSERT_NOT_NULL(shader);
  ASSERT_NOT_NULL(matrix);
  ASSERT_GREATER_THAN_EQUAL_TO(uniform, 0);
  glUniformMatrix4fv(uniform, 1, GL_FALSE, matrix->internalMatrix[0]);
}

void shaderUseColor(
  shaderprogram_t *shader, shaderuniform_t uniform, pixel_t color
) {
  ASSERT_NOT_NULL(shader);
  ASSERT_GREATER_THAN_EQUAL_TO(uniform, 0);
  glUniform4f(uniform,
    (float)color.r / 255.0f,
    (float)color.g / 255.0f,
    (float)color.b / 255.0f,
    (float)color.a / 255.0f 
  );
}

shaderuniform_t shaderGetUniformByName(shaderprogram_t *sp, const char *name) {
  ASSERT_NOT_NULL(sp);
  ASSERT_NOT_NULL(name);
  return glGetUniformLocation(sp->shaderProgram, name);
}